{****************************************************}
{ Hashed List implementation unit for Borland Delphi }
{ Copyright  2002 by Dimka Maslov                   }
{ e-mail:   mainbox@endimus.com                      }
{ web-site: http://endimus.com                       }
{****************************************************}

unit HshdList;

{ This unit is a realization of hashed lists also called
  as hash-arrays or associated arrays for Borland Delphi.

  The hashed list is an array of items where each item
  can be idenified by the unique name. (Not by the integer
  number). Hashing is used to increase the speed to access
  items in list. Prior to including an item into a list
  the hashsum (MD5 in this example) of a name is calculated.
  And new item is placed into a list to keep the order of
  hashsums. Searching an item by specified name compares only
  sorted hashsums, not unsorted string names.

  To customize list it is need to code one class
  descentant and create an instance of list specifying
  the name of items class only!
 }

interface

uses SysUtils, Classes, Dim, MD5;

{
 The Dim.pas unit may be downloaded at http://downloads.endimus.com/dimpas.zip
 The MD5.pas unit may be downloaded at http://downloads.endimus.com/md5.zip
}

type

  THashedList = class;

  { THashedListItem class realizes the minimum functionality
    for a hashed list item. Class of custom hashed list items
    must be a descendant of this class. }
  THashedListItem = class (TPersistent)
  private
    FName: TString;
    FStatus: Integer;
    FList: THashedList;
    FHash: TMD5Digest;
    FTag: Integer;
    procedure SetName(Value: TString);
    constructor Create(AName: TString; AList: THashedList);
  protected
  { The AssignTo method is called by TPersistent.Assign
    method to copy data to another object. This method
    must be overridden to copy any custom data of
    descendant class items. }
    procedure AssignTo(Dest: TPersistent); override;
  public
  { The Name property is used to identify an item.
    Identification of items is case-sensitive.
    Convert a string into lower-case or upper-case
    prior to assigning names to make names case-
    insensitive. }
    property Name: TString read FName write SetName;
  { The Hash property is automatically calculated after
    assigning the Name property and contains the MD5
    hashsum of the Name. }
    property Hash: TMD5Digest read FHash;
  { The Tag property has user defined destination. }
    property Tag: Integer read FTag write FTag;
  end;

  THashedListItemClass = class of THashedListItem;

  PHashedListItems = ^THashedListItems;
  THashedListItems = array[0..0] of THashedListItem;

  EHashedList = class (Exception);

  { The THashedList class realizes the self-sorted
    list of items. Each item added to a list is
    placed to keep the order of MD5 hashsum of items
    names.}
  THashedList = class (TObject)
  private
    FItemClass: THashedListItemClass;
    FCount: Integer;
    FCapacity: Integer;
    FBuffer: PHashedListItems;
    procedure ItemChanged(Item: THashedListItem; Index: Integer);
    procedure ItemInserted(Item: THashedListItem);
    function Get(AName: TString): THashedListItem;
    procedure Search(const Hash: TMD5Digest; out Item: THashedListItem;
      out Index: Integer; out Cmp: Integer);
    procedure Delete(Index: Integer); overload;
    procedure Insert(Item: THashedListItem; Index: Integer);
    procedure Grow;
    procedure DoneList;
    function CompareMD5(const Hash1, Hash2: TMD5Digest): LongInt;
  public
  { The Add method creates a new item and places it into a list.
    This function raises an exception if the list has the item with
    specified name. The class or generated items is specified by
    the ItemClass property. }
    function Add(AName: TString): THashedListItem;
  { The Delete method removes an item from a list. This procedure
    raises an exception if there is no item with specified name. }
    procedure Delete(AName: TString); overload;
  { The Items default property implements access to list items by names.
    This property is nil if there is no item with specified
    name. }
    property Items[AName: TString]: THashedListItem read Get; default;
  { The ItemByIndex function provides access to list items by indices
    in order of sorting by Add method. This method is useful to reach
    every item of list. }
    function ItemByIndex(Index: Integer): THashedListItem;
  { The IndexOfItem function returns the index of the item with
    specified name or -1 if there is no item found. }
    function IndexOfItem(Name: TString): Integer;
  { The Count property is the number of items of a list. }
    property Count: Integer read FCount;
  { The ItemClass property specifies the class of list items.
    The Add function creates items of a class specified by
    this property. This property is assigned by constructor. }
    property ItemClass: THashedListItemClass read FItemClass;
  { The Create constructor receives the one argument that
    specifies the class of items. Since that there is no
    need to code descendant list class for each items class. }
    constructor Create(AItemClass: THashedListItemClass);

    destructor Destroy; override;
  end;

  { THashedStringsItem and THashedStrings classes demonstrates an
    example for using THashedListItem and THashedList classes. }


  { THashedStringsItem class objects are used to store data of
    THashedStrings objects. }

  THashedStringsItem = class (THashedListItem)
  private
    FValue: TString;
    FObject: TObject;
    procedure SetValue(const Value: TString);
    procedure SetObject(const Value: TObject);
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
  { The Value property represents strings which accessed
    by the THashedStrings.Strings property. }
    property Value: TString read FValue write SetValue;
  { The Obj property represents objects which acceessed
    by the THashedStrings.Objects property. }
    property Obj: TObject read FObject write SetObject;
  end;

  { THashedStrings implements a hash-array of strings. Unlike
    THashedList, names of strings of this class objects is not
    case-sensitive. }
  THashedStrings = class (THashedList)
  private
    function GetObjects(Name: TString): TObject;
    function GetStrings(Name: TString): TString;
    procedure SetObjects(Name: TString; const Value: TObject);
    procedure SetStrings(Name: TString; const Value: TString);
    function GetItem(Name: TString): THashedStringsItem;
    function GetItems(Name: TString): THashedStringsItem;
    procedure SetItems(Name: TString; const Value: THashedStringsItem);
  public
  { The Strings property reads, writes and adds strings with specified name to
    the list. References to this property in the right side of an expresson
    read strings or raise exceptions for unknown names. References to this
    property in the left side of an expression assign existing strings or
    includes new strings when in a list there is not a string with specified
    name. }
    property Strings[Name: TString]: TString read GetStrings write SetStrings; default;
  { Each string in a list has the accociated object for user defined purposes.
    The Objects property provides access to those objects. References to this
    property act like references to the Strings property. }
    property Objects[Name: TString]: TObject read GetObjects write SetObjects;
  { The Items property provides access to THashedStringsItem class objects
    like Strings or Objects properties do.  }
    property Items[Name: TString]: THashedStringsItem read GetItems write SetItems;
  { The ItemByIndex function calls inherited casts result to
    THashedStringsItem class. }
    function ItemByIndex(Index: Integer): THashedStringsItem;
    constructor Create;
  end;


resourcestring
  // These strings may be localized
  SHashedListDuplicates = 'List item already exists - %s';
  SHashedListItemMissing = 'List item not found - %s';

implementation

uses Consts;

{ THashedListItem }

procedure THashedListItem.AssignTo(Dest: TPersistent);
begin
 if Dest is THashedListItem
  then THashedListItem(Dest).FTag := Self.FTag else inherited;
end;

constructor THashedListItem.Create(AName: TString; AList: THashedList);
begin
 FStatus:=-1;
 FList:=AList;
 SetName(AName);
end;

procedure THashedListItem.SetName(Value: TString);
var
 OldName: TString;
 OldHash: TMD5Digest;
 OutItem: THashedListItem;
 Index: Integer;
 Cmp: Integer;
begin
 if FStatus < 0 then begin
  FName:=Value;
  FHash:=MD5String(FName);
  FList.ItemInserted(Self);
 end else begin
  OldName:=FName;
  OldHash:=FHash;
  try
   FList.Search(FHash, OutItem, Index, Cmp);
   if Integer(OutItem) = Integer(Self) then begin
    FName:=Value;
    FHash:=MD5String(FName);
    FList.ItemChanged(Self, Index);
   end;
  except
   FName:=OldName;
   FHash:=OldHash;
   FList.ItemInserted(Self);
   raise;
  end;
 end;
end;

{ THashedList }

function THashedList.Add(AName: TString): THashedListItem;
begin
 Result:=ItemClass.Create(AName, Self);
end;

function THashedList.CompareMD5(const Hash1, Hash2: TMD5Digest): LongInt;
begin
 Result:=Hash1.A - Hash2.A;
 if Result <> 0 then Exit;
 Result:=Hash1.B - Hash2.B;
 if Result <> 0 then Exit;
 Result:=Hash1.C - Hash2.C;
 if Result <> 0 then Exit;
 Result:=Hash1.D - Hash2.D;
end;

constructor THashedList.Create(AItemClass: THashedListItemClass);
begin
 FItemClass:=AItemClass;
end;

procedure THashedList.Delete(Index: Integer);
begin
 if Index < FCount then MoveMem(FBuffer^[Index + 1], FBuffer^[Index],
     (FCount - Index) * SizeOf(THashedListItem));
 Dec(FCount);
 FBuffer^[FCount] := nil;
end;

procedure THashedList.Delete(AName: TString);
var
 Item: THashedListItem;
 Index: Integer;
 Cmp: Integer;
begin
 Search(MD5String(AName), Item, Index, Cmp);
 if Assigned(Item) then begin
  Delete(Index);
  Item.Free;
 end else raise EHashedList.CreateFmt(SHashedListItemMissing, [AName]);
end;

destructor THashedList.Destroy;
begin
  DoneList;
  inherited;
end;

procedure THashedList.DoneList;
var
 i: Integer;
begin
 for i:=0 to FCount - 1 do FreeAndNil(FBuffer^[i]);
 DeallocateMem(FBuffer);
end;

function THashedList.Get(AName: TString): THashedListItem;
var
 Index: Integer;
 Cmp: Integer;
begin
 Search(MD5String(AName), Result, Index, Cmp);
end;

procedure THashedList.Grow;
var
 Delta: Integer;
begin
 if FCapacity > 64 then Delta := FCapacity shr 2 else
  if FCapacity > 8 then Delta := 16 else Delta := 4;
 Inc(FCapacity, Delta);
 ReallocateMem(FBuffer, FCapacity, SizeOf(THashedListItem));
 ClearMem( IncPtr(FBuffer, FCount*SizeOf(THashedListItem))^,
  (FCapacity-FCount)*SizeOf(THashedListItem) );
end;

function THashedList.IndexOfItem(Name: TString): Integer;
var
 Index: Integer;
 Item: THashedListItem;
 Cmp: Integer;
begin
 Search(MD5String(Name), Item, Index, Cmp);
 if Assigned(Item) then Result:=Index else Result:=-1;
end;

procedure THashedList.Insert(Item: THashedListItem; Index: Integer);
begin
 if FCount = FCapacity then Grow;
 if Index < FCount then  MoveMem(FBuffer^[Index], FBuffer^[Index + 1],
  (FCount - Index) * SizeOf(THashedListItem));
 FBuffer^[Index] := Item;
 Item.FStatus := 1;
 Inc(FCount);
end;

function THashedList.ItemByIndex(Index: Integer): THashedListItem;
begin
 if (Index>=0) and (Index<FCount) then Result:=FBuffer^[Index] else Result:=nil;
end;

procedure THashedList.ItemChanged(Item: THashedListItem; Index: Integer);
begin
 Delete(Index);
 Item.FStatus:=0;
 ItemInserted(Item);
end;

procedure THashedList.ItemInserted(Item: THashedListItem);
var
 Index: Integer;
 OutItem: THashedListItem;
 Cmp: Integer;
begin
 Search(Item.FHash, OutItem, Index, Cmp);
 if Cmp > 0 then Inc(Index);
 if Assigned(OutItem) then
  raise EHashedList.CreateFmt(SHashedListDuplicates, [Item.Name]);
 Insert(Item, Index);
end;

procedure THashedList.Search(const Hash: TMD5Digest; out Item: THashedListItem;
  out Index: Integer; out Cmp: Integer);
var
 IndexLow, IndexHigh: Integer;
begin
 if FCount = 0 then begin
  Item:=nil;
  Index:=0;
  Cmp:=0;
 end else begin
  IndexLow:=0;
  IndexHigh:=FCount - 1;
  Index:=0;
  Item := nil;
  repeat
   Index := (IndexLow + IndexHigh) shr 1;
   Cmp:=CompareMD5(Hash, FBuffer^[Index].FHash);
   if Cmp = 0 then begin
    Item:=FBuffer^[Index];
    Break;
   end;
   if Cmp < 0 then IndexHigh := Index - 1 else IndexLow := Index + 1;
  until IndexLow > IndexHigh;
 end;
end;

{ THashedStringsItem }

procedure THashedStringsItem.AssignTo(Dest: TPersistent);
begin
 inherited;
 if Dest is THashedStringsItem then begin
  THashedStringsItem(Dest).Value := Self.Value;
  THashedStringsItem(Dest).FObject:= Self.FObject;
 end else
  raise EConvertError.CreateFmt(SAssignError, [ClassName, Dest.ClassName]);
end;

procedure THashedStringsItem.SetObject(const Value: TObject);
begin
  FObject := Value;
end;

procedure THashedStringsItem.SetValue(const Value: TString);
begin
  FValue := Value;
end;

{ THashedStrings }

constructor THashedStrings.Create;
begin
 inherited Create(THashedStringsItem);
end;

function THashedStrings.GetItem(Name: TString): THashedStringsItem;
begin
 Name:=DnString(Name);
 Result:=inherited Items[Name] as THashedStringsItem;
 if not Assigned(Result) then
  Result:=inherited Add(Name) as THashedStringsItem;
end;

function THashedStrings.GetItems(Name: TString): THashedStringsItem;
var
 Item: THashedListItem;
begin
 Item:= inherited Items[DnString(Name)];
 if Assigned(Item) then Result:=Item as THashedStringsItem
 else raise EHashedList.CreateFmt(SHashedListItemMissing, [Name]);
end;

function THashedStrings.GetObjects(Name: TString): TObject;
begin
 Result:=Items[Name].Obj;
end;

function THashedStrings.GetStrings(Name: TString): TString;
begin
 Result:=Items[Name].Value;
end;

function THashedStrings.ItemByIndex(Index: Integer): THashedStringsItem;
begin
 Result:=inherited ItemByIndex(Index) as THashedStringsItem;
end;

procedure THashedStrings.SetItems(Name: TString;
  const Value: THashedStringsItem);
begin
 GetItem(Name).Assign(Value);
end;

procedure THashedStrings.SetObjects(Name: TString; const Value: TObject);
begin
 GetItem(Name).Obj:=Value;
end;

procedure THashedStrings.SetStrings(Name: TString; const Value: TString);
begin
 GetItem(Name).Value:=Value;
end;

end.
